@@ -3,9 +3,8 @@ use warnings;
 use strict;
 use JSON::Create;
 my $jc = JSON::Create->new ();
-$jc->validate (1);
 $jc->type_handler (sub {
-		       return ('"a"', '"b"', '"c"');
+		       return (1, 2, 3);
 		   });
 print $jc->run ({ x => *STDOUT }); 
 
