@@ -28,7 +28,6 @@ typedef enum {
     json_create_bad_floating_format,
     /* */
     json_create_unicode_bad_utf8,
-    json_create_code_ref_found,
     /* User's routine returned invalid stuff. */
     json_create_invalid_user_json,
 }
@@ -55,14 +54,10 @@ typedef struct json_create {
     /* Handlers for objects and booleans. If there are no handlers,
        this is zero (a NULL pointer). */
     HV * handlers;
-    /* Code reference handler. */
-    SV * code_ref_handler;
+    /* User reference handler. */
+    SV * type_handler;
     /* Do any of the SVs have a Unicode flag? */
     unsigned int unicode : 1;
-    /* Does the SV we're currently looking at have a Unicode flag? */
-    unsigned int unicode_now : 1;
-    /* Did we see non-Unicode and non-ASCII bytes? */
-    unsigned int non_unicode : 1;
     /* Should we convert / into \/? */
     unsigned int escape_slash : 1;
     /* Should Unicode be upper case? */
@@ -71,8 +66,10 @@ typedef struct json_create {
     unsigned int unicode_escape_all : 1;
     /* Should we validate user-defined JSON? */
     unsigned int validate : 1;
-    /* */
+    /* Do not escape U+2028 and U+2029. */
     unsigned int no_javascript_safe : 1;
+    /* Make errors fatal. */
+    unsigned int fatal_errors : 1;
 }
 json_create_t;
 
@@ -89,16 +86,6 @@ json_create_t;
 	CALL (json_create_buffer_fill (jc));	\
     }
 
-static int
-perl_error_handler (const char * file, int line_number, const char * msg, ...)
-{
-    va_list args;
-    va_start (args, msg);
-    vcroak (msg, & args);
-    va_end (args);
-    return 0;
-}
-
 /* Print an error to stderr. */
 
 static int
@@ -123,7 +110,6 @@ static int (* json_create_error_handler) (const char * file, int line_number, co
 	switch (status) {					\
 	    /* These exceptions indicate a user error. */	\
 	case json_create_unknown_type:				\
-	case json_create_code_ref_found:			\
 	case json_create_unicode_bad_utf8:			\
 	case json_create_invalid_user_json:			\
 	    break;						\
@@ -147,6 +133,20 @@ static int (* json_create_error_handler) (const char * file, int line_number, co
 	}								\
     }
 
+static void
+json_create_user_message (json_create_t * jc, json_create_status_t status, const char * format, ...)
+{
+    va_list a;
+    /* Check the status. */
+    va_start (a, format);
+    if (jc->fatal_errors) {
+	vcroak (format, & a);
+    }
+    else {
+	vwarn (format, & a);
+    }
+}
+
 /* Everything else in this file is ordered from callee at the top to
    caller at the bottom, but because of the recursion as we look at
    JSON values within arrays or hashes, we need to forward-declare
@@ -377,7 +377,7 @@ json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN k
 		/* Three byte case. */
 		if (input[1] < 0x80 || input[1] > 0xBF ||
 		    input[2] < 0x80 || input[2] > 0xBF) {
-		    warn ("Invalid UTF-8");
+		    json_create_user_message (jc, json_create_unicode_bad_utf8, "Invalid UTF-8");
 		    return json_create_unicode_bad_utf8;
 		}
 		if (! jc->no_javascript_safe &&
@@ -403,7 +403,7 @@ json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN k
 	    else if ((c & 0xC0) == 0xC0) {
 		/* Two byte case. */
 		if (input[1] < 0x80 || input[1] > 0xBF) {
-		    warn ("Invalid UTF-8");
+		    json_create_user_message (jc, json_create_unicode_bad_utf8, "Invalid UTF-8");
 		    return json_create_unicode_bad_utf8;
 		}
 		if (jc->unicode_escape_all) {
@@ -418,12 +418,9 @@ json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN k
 		i += 2;
 	    }
 	    else {
-		warn ("Invalid UTF-8");
+		json_create_user_message (jc, json_create_unicode_bad_utf8, "Invalid UTF-8");
 		return json_create_unicode_bad_utf8;
 	    }
-	    if (! jc->unicode_now) {
-		jc->non_unicode = 1;
-	    }
 	}
 	if (j == i) {
 	    fprintf (stderr, "Failed to increment %d %d\n", i, j);
@@ -442,22 +439,14 @@ json_create_add_string (json_create_t * jc, SV * input)
     char * istring;
     STRLEN ilength;
     /*
-      "jc->unicode_now" is true (equals 1) if Perl says that "input"
-      contains a "SvUTF8" scalar, and false (equals 0) if Perl does
-      not say that "input" contains a "SvUTF8" scalar.
-
       "jc->unicode" is true if Perl says that anything in the whole of
       the input to "json_create" is a "SvUTF8" scalar.
     */
-    jc->unicode_now = 0;
     istring = SvPV (input, ilength);
     if (SvUTF8 (input)) {
 	/* We have to force everything in the whole output to
 	   Unicode. */
 	jc->unicode = 1;
-	/* Flag that this string is supposed to be Unicode to the
-	   upstream. */
-	jc->unicode_now = 1;
     }
     /* Backtrace fall through, remember to check the caller's line. */
     return json_create_add_key_len (jc, (unsigned char *) istring,
@@ -706,16 +695,11 @@ json_create_add_object (json_create_t * jc, HV * input_hv)
 	HE * he;
 
 	/* Get the information from the hash. */
-
-	/* The key is not unicode unless Perl tells us it's
-	   unicode. */
-	jc->unicode_now = 0;
 	/* The following is necessary because "hv_iternextsv" doesn't
 	   tell us whether the key is "SvUTF8" or not. */
 	he = hv_iternext (input_hv);
 	if (HeUTF8 (he)) {
 	    jc->unicode = 1;
-	    jc->unicode_now = 1;
 	}
 	key = hv_iterkey (he, & keylen);
 	value = hv_iterval (input_hv, he);
@@ -756,44 +740,6 @@ json_create_add_array (json_create_t * jc, AV * av)
     return json_create_ok;
 }
 
-/*
-
-Copied from
-
-https://metacpan.org/source/TOBYINK/match-simple-XS-0.001/XS.xs#L11
-
-via
-
-http://grep.cpan.me/?q=SvRX
-
-*/
-
-#ifndef SvRXOK
- 
-#define SvRXOK(sv) is_regexp(aTHX_ sv)
- 
-static INLINE int
-is_regexp (pTHX_ SV* sv)
-{
-    SV* tmpsv;
-         
-    if (SvMAGICAL (sv)) {
-	mg_get (sv);
-    }
-         
-    if (SvROK (sv)
-        && (tmpsv = (SV*) SvRV (sv))
-        && SvTYPE (tmpsv) == SVt_PVMG 
-        && (mg_find (tmpsv, PERL_MAGIC_qr))) {
-	return TRUE;
-    }
-    return FALSE;
-}
- 
-#endif
-
-/* <-- End of Toby Inkster contribution. Thank you. */
-
 #define UNKNOWN_TYPE_FAIL(t)				\
     if (JCEH) {						\
 	(*JCEH) (__FILE__, __LINE__,			\
@@ -822,8 +768,9 @@ json_create_validate_user_json (json_create_t * jc, SV * json)
 	return json_create_ok;
     }
     if (SvOK (error) && SvCUR (error) > 0) {
-	warn ("JSON::Parse::assert_valid_json failed for '%s': %s",
-	      SvPV_nolen (json), SvPV_nolen (error));
+	json_create_user_message (jc, json_create_invalid_user_json,
+				  "JSON::Parse::assert_valid_json failed for '%s': %s",
+				  SvPV_nolen (json), SvPV_nolen (error));
 	return json_create_invalid_user_json;
     }
     return json_create_ok;
@@ -847,11 +794,7 @@ json_create_call_to_json (json_create_t * jc, SV * cv, SV * r)
     //https://metacpan.org/source/AMBS/Math-GSL-0.35/swig/gsl_typemaps.i#L482
     XPUSHs (sv_2mortal(newRV(r)));
     PUTBACK;
-    count = call_sv (cv, 0);
-    if (count != 1) {
-	warn ("Wrong number of arguments %ld from user callback: should be 1",
-	      count);
-    }
+    call_sv (cv, 0);
     json = POPs;
     SvREFCNT_inc (json);
     FREETMPS;
@@ -865,6 +808,20 @@ json_create_call_to_json (json_create_t * jc, SV * cv, SV * r)
     return json_create_ok;
 }
 
+static INLINE json_create_status_t
+json_create_handle_unknown_type (json_create_t * jc, SV * r)
+{
+    if (jc->type_handler) {
+	CALL (json_create_call_to_json (jc, jc->type_handler, r));
+    }
+    else {
+	json_create_user_message (jc, json_create_unknown_type,
+				  "Input's type cannot be serialized to JSON");
+	return json_create_unknown_type;
+    }
+    return json_create_ok;
+}
+
 static INLINE json_create_status_t
 json_create_handle_ref (json_create_t * jc, SV * input)
 {
@@ -872,10 +829,6 @@ json_create_handle_ref (json_create_t * jc, SV * input)
     SV * r;
     r = SvRV (input);
     t = SvTYPE (r);
-    /* First try a switch for the types which have been in Perl
-       for a while. We can't add the case of SVt_REGEXP here since
-       it's not present in some older Perls, so we test for
-       regexes in the default: case at the bottom. */
     switch (t) {
     case SVt_PVAV:
 	CALL (json_create_add_array (jc, (AV *) r));
@@ -908,27 +861,8 @@ json_create_handle_ref (json_create_t * jc, SV * input)
 	CALL (json_create_add_string (jc, r));
 	break;
 
-    case SVt_PVCV:
-	if (jc->code_ref_handler) {
-	    CALL (json_create_call_to_json (jc, jc->code_ref_handler, r));
-	}
-	else {
-	    warn ("Code reference cannot be serialized to JSON");
-	    return json_create_code_ref_found;
-	}
-	break;
-
     default:
-	/* Test for regex, possibly using the Toby Inkster code
-	   above. */
-	if (SvRXOK (r)) {
-	    /* Use it as a string. */
-	    CALL (json_create_add_string (jc, r));
-	}
-	else {
-	    warn ("Input's type cannot be serialized to JSON");
-	    return json_create_unknown_type;
-	}
+	CALL (json_create_handle_unknown_type (jc, r));
     }
     return json_create_ok;
 }
@@ -1093,7 +1027,7 @@ json_create_recursively (json_create_t * jc, SV * input)
 	    break;
 	    
 	default:
-	    UNKNOWN_TYPE_FAIL (t);
+	    CALL (json_create_handle_unknown_type (jc, r));
 	}
     }
     return json_create_ok;
@@ -1139,8 +1073,6 @@ json_create_run (json_create_t * jc, SV * input)
     jc->output = 0;
     /* Not Unicode. */
     jc->unicode = 0;
-    /* So far we have not seen any non-Unicode bytes over 0x80. */
-    jc->non_unicode = 0;
 
     /* Unleash the dogs. */
     FINALCALL (json_create_recursively (jc, input));
@@ -1148,24 +1080,6 @@ json_create_run (json_create_t * jc, SV * input)
     FINALCALL (json_create_buffer_fill (jc));
 
     if (jc->unicode) {
-
-    /* At least one of the SVs was Unicoded, so switch on the Unicode
-       flag in jc->output.
-
-       We also checked that there was nothing with a non-ASCII byte in
-       an SV not marked as Unicode, so we are now going to trust that
-       the user did not send insane inputs. */
-
-	if (jc->non_unicode) {
-
-	    /* If there was something with a non-ASCII byte not marked
-	       as Unicode, we're going to just refuse to encode it. */
-
-	    warn ("Mixed multibyte and binary inputs, "
-		  "refusing to encode to JSON");
-	    SvREFCNT_dec (jc->output);
-	    return & PL_sv_undef;
-	}
 	SvUTF8_on (jc->output);
     }
 
@@ -1195,7 +1109,7 @@ json_create (SV * input)
 
     jc.unicode_escape_all = 0;
     jc.handlers = 0;
-    jc.code_ref_handler = 0;
+    jc.type_handler = 0;
 
     return json_create_run (& jc, input);
 }
@@ -1208,7 +1122,7 @@ json_create_new (json_create_t ** jc_ptr)
     jc->n_mallocs = 0;
     jc->n_mallocs++;
     jc->fformat = 0;
-    jc->code_ref_handler = 0;
+    jc->type_handler = 0;
     jc->handlers = 0;
     * jc_ptr = jc;
     return json_create_ok;
@@ -1264,11 +1178,11 @@ json_create_remove_handlers (json_create_t * jc)
 }
 
 static json_create_status_t
-json_create_remove_code_handler (json_create_t * jc)
+json_create_remove_type_handler (json_create_t * jc)
 {
-    if (jc->code_ref_handler) {
-	SvREFCNT_dec (jc->code_ref_handler);
-	jc->code_ref_handler = 0;
+    if (jc->type_handler) {
+	SvREFCNT_dec (jc->type_handler);
+	jc->type_handler = 0;
 	jc->n_mallocs--;
     }
     return json_create_ok;
@@ -1279,7 +1193,7 @@ json_create_free (json_create_t * jc)
 {
     CALL (json_create_free_fformat (jc));
     CALL (json_create_remove_handlers (jc));
-    CALL (json_create_remove_code_handler (jc));
+    CALL (json_create_remove_type_handler (jc));
 
     /* Finished, check we have no leaks before freeing. */
 
